open B8_types

let argc = Array.length Sys.argv
let time = Unix.gettimeofday ()
let delay = ref (-1)
let dump_doc = ref false
let dump_tdoc = ref false
let use_opts = ref []
let dump_size = ref false
let quiet = ref false
let save_temps = ref false
let dump_opts = ref false
let valid_opts = ref true
let valid_result = ref false
let dump_exn = ref false
let disabled = ref []
let dump_inter = ref false
let output_file = ref ""
let input_file = ref ""
  
let small_string _ =
  let n = 2 + (Random.int 10) in
  let s = String.create n in
  for i = 0 to n - 1 do
    s.[i] <- (match Random.int 3 with
        0 -> ' '
      | 1 -> char_of_int (65 + Random.int 26)
      | _ -> char_of_int (97 + Random.int 26))
  done;
  s
  
let rec do_gen size =
  let expr =
    if size = 0 then
      Tstring (small_string ())
    else
    match Random.int 30 with
    | 10 -> Ttag_B (do_gen (size - 1))
    | 11 -> Ttag_EM (do_gen (size - 1))
    | 12 -> Ttag_I  (do_gen (size - 1))
    | 13 -> Ttag_PL (do_gen (size - 1))
    | 14 -> Ttag_S  (do_gen (size - 1))
    | 15 -> Ttag_TT (do_gen (size - 1))
    | 16 -> Ttag_U  (do_gen (size - 1))
    | 17 | 26 | 27 | 28 | 29 -> Tlist ((do_gen (size - 1) :: (list_gen size)))
    | n -> 
        if n < 18 then
          Ttag_size (do_gen (size - 1), B8_utils.size_of_int n)
        else
          Ttag_color (do_gen (size - 1), B8_utils.color_of_int (n-18))
  in    
  { expr = expr; prop = prop_root; }
  
and list_gen size =
  if Random.int 10 > 5 then
    (do_gen (size - 1)) :: (list_gen size)
  else
    [ { expr = Tstring (small_string ()); prop = prop_root; }]
    
let do_gen size =
  let _ = Random.self_init () in
  let tdoc = do_gen size in
  B8_emit.emit_oc tdoc stdout;
  print_newline ();
  flush stdout;
  exit 0
  
let do_quiet _ = 
  quiet := true; dump_size := true

let do_S _ =
  do_quiet (); save_temps := true

let tdoc_of_ic ic =
  let lexbuf = Lexing.from_channel ic in
  let doc = B8_parser.main  B8_lexer.token lexbuf in
  let tdoc = B8_typedoc.typedoc doc in
  tdoc

let set_valid _ =
  valid_opts := true;
  valid_result := true

let do_debug _ =
  do_S (); dump_opts := true;
  set_valid ();
  dump_exn := true

let do_disable s =
  Printf.printf "DISABLE %s" s; print_newline ();
  disabled := s :: !disabled
  
let do_valid f =
  let ic = open_in f in
  Printf.printf "Reading %s" f; print_newline ();
  let tdoc1 = tdoc_of_ic ic in
  Printf.printf "Reading stdin"; print_newline ();
  let tdoc2 = tdoc_of_ic stdin in
  Printf.printf "Check validity"; print_newline ();
  if B8_valid.validate tdoc1 tdoc2 then
    (Printf.printf "Both files are equivalent"; print_newline (); exit 0)
  else
    (Printf.printf "Files are different"; print_newline (); 
      B8_emit.escape := true;
      B8_emit.emit tdoc1 "nequiv1.sml";
      B8_emit.emit tdoc2 "nequiv2.sml";
      exit 1)

let _ =
  Arg.parse [
    "-gen", Arg.Int do_gen, " <n> : generate a file with <n> nodes";
    "-dump_inter", Arg.Set dump_inter, " : dump intermediate results"; 
    "-dump_doc", Arg.Set dump_doc,   " : dump doc after parsing ";
    "-dump_tdoc", Arg.Set dump_tdoc, " : dump tdoc after typedoc ";
    "-dump_size", Arg.Set dump_size, " : print input and output sizes ";
    "-quiet", Arg.Unit do_quiet, " : don't output result + -dump_size";
    "-S", Arg.Unit do_S, " : don't remove temporary files + -quiet";
    "-dump_opts", Arg.Set dump_opts, " : dump opt names";
    "-debug", Arg.Unit do_debug, " : -S -dump_opts";
    "-g", Arg.Unit do_debug, " : -S -dump_opts";
    "-valid", Arg.String do_valid, " <filename>: compare stdin with <filename>";
    "-v", Arg.Set valid_result, " : check validity of result";
    "-vv", Arg.Set valid_opts, " : check validity after each opt";
    "-vvv", Arg.Unit set_valid, " : -v -vv";
    "-disable", Arg.String do_disable, " <opt>: disable <opt>";
    "-o", Arg.String ((:=) output_file), " <filename>: set output file";
    "-f", Arg.String ((:=) input_file), " <filename>: set input file";
  ] (fun s -> 
      if !delay < 0 then 
        delay := int_of_string s
      else
        use_opts := s :: !use_opts) "contest delay [opts]"
  
let use_opts = List.rev !use_opts
let delay = !delay

(* Computes the time left for optimizations. Should be dependent on 
the input size ! *)
let opt_time delay =
  min (delay - 20) ((delay * 9) / 10) (* 10 seconds kept for output *)
  
let output_result tdoc =
  let oc = if !output_file = "" then stdout else open_out !output_file in
(* output result *)
  let filename = Filename.temp_file "contest" ".tmp" in
  B8_emit.emit tdoc filename;
  let st = Unix.stat filename in
  if !dump_size then begin
      Printf.printf "Final size: %d" st.Unix.st_size;
      print_newline ();
    end;
  if not !quiet || !output_file <> "" then begin
      let ic = open_in filename in
      let s = String.create 8192 in
      let rec iter () =
        let n = input ic s 0 8129 in
        if n > 0 then
          (output oc s 0 n; iter ())
      in
      iter ();
      close_in ic;
      flush stdout;
      if !output_file <> "" then close_out oc;
    end;
  if not !save_temps then Sys.remove filename else
    (Printf.printf "Result: %s" filename; print_newline ());
  exit 0

  
let _ =
  let ic = if !input_file = "" then stdin else open_in !input_file in
  let lexbuf = Lexing.from_channel ic in
  let doc = B8_parser.main  B8_lexer.token lexbuf in
  if !input_file <> "" then  close_in ic;
  if !dump_size then begin
      Printf.printf "Initial size: %d" (Lexing.lexeme_end lexbuf);
      print_newline ();
    end;
  if !dump_doc then begin
      Printf.printf "Initial text:"; print_newline ();
      B8_dump.dump_doc doc;
      print_newline ();
    end;
  let tdoc = B8_typedoc.typedoc doc in
  if !dump_tdoc then begin
      Printf.printf "Final text:"; print_newline ();
      B8_dump.dump_tdoc tdoc;
      print_newline ();
    end;
  
  let first_doc = tdoc in
  let tdoc = ref tdoc in
  let tdoc_size = ref (B8_valid.size !tdoc) in
  if !quiet then begin
      Printf.printf "Computed initial size %d" !tdoc_size;
      print_newline ();
    end;
  if delay >= 0 then begin
      ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle (fun _ -> 
              if !quiet then begin
                  Printf.printf "TIMEOUT"; print_newline ();
                end;
              output_result !tdoc)));
      ignore (Unix.setitimer  Unix.ITIMER_REAL { 
          Unix.it_value = float_of_int (opt_time delay);
          Unix.it_interval = 0.0 })
    end;
(* optimisations *)
  if use_opts = [] then
    
    let rec compute_opts list indoc insize =
      match list with
        [] -> ()
      | f :: tail ->
          let (indoc, insize) = 
            try  
              let td = f indoc in
              let td = B8_typedoc.retype_tdoc td in
              let new_size = B8_valid.size td in
              if !quiet then begin
                  Printf.printf "OPT output size: %d" new_size;
                  print_newline ();
                end;
              if new_size > insize then
                begin
                  if !quiet then begin
                      Printf.printf "OPT output bigger document !!!";
                      print_newline ();
                    end;
                  (indoc, insize)
                end
              else
              if not !valid_opts || B8_valid.validate td first_doc then
                begin
                  if new_size < !tdoc_size then
                    (tdoc := td; tdoc_size := new_size);
                  (td, new_size)
                end
              else begin
                  if !quiet then begin
                      Printf.printf "OPT output incorrect result !!!";
                      print_newline ();
                    end;
                  (indoc, insize)
                end
            
            
            with e -> 
                if !dump_exn then
                  (Printf.printf "exception %s" (
                      Printexc.to_string e); print_newline ());
                (indoc, insize)
          in
          compute_opts tail indoc insize
    in
    
    begin
      let first_opts = [
          B8_vide.garbage ; 
          B8_cro.merge
        ] in
      compute_opts first_opts !tdoc !tdoc_size;
      
      let tdoc2 = !tdoc in
      let tdoc_size2 = !tdoc_size in
      
      if !quiet then begin
          Printf.printf "First OPT done";
          print_newline ();
        end;
      
      let first_stack = [ 
          B8_cro.redondant; 
          B8_cro.subflatten;
          B8_colornesting.nesting_flat;
          B8_vide.garbage ; 
          B8_cro.redondant; 
          B8_cro.subflatten;
          B8_colornesting.nesting_flat;
        ] in
      ignore (compute_opts first_stack tdoc2 tdoc_size2);
      Gc.compact ();
      
      if !quiet then begin
          Printf.printf "First stack done";
          print_newline ();
        end;
      
      if !tdoc_size < 1000000 then begin
          let second_stack = [
              B8_cro.flat_and_build; 
              B8_flat.flat_opt;
            ] in
          ignore (compute_opts second_stack tdoc2 tdoc_size2);
          Gc.compact ();
          if !quiet then begin
              Printf.printf "Second stack done";
              print_newline ();
            end;
        end;      
      
      let final_stack = [
          B8_cro.subflatten;
          B8_vide.garbage;
          B8_colornesting.nesting_flat;
          B8_cro.redondant;
          B8_cro.merge;
          B8_cro.simplify;
          B8_colornesting.nesting;
        ] in
      
      let previous_size = ref (!tdoc_size+1) in
      while !tdoc_size < !previous_size do
        previous_size := !tdoc_size;
        ignore (compute_opts final_stack !tdoc !tdoc_size);
        Gc.compact ();
        if !quiet then begin
            Printf.printf "Done final stack";
            print_newline ();
          end;
        
      done

    end
    
    
    (*
    let opts = Sort.list (fun (_,(_,i)) (_,(_,j)) -> i < j) !B8_types.opts in
    List.iter (fun (s,(f,_)) ->
        if List.mem s !disabled then
          (Printf.printf "OPT: %s DISABLED" s; print_newline ())
        else
        try 
          if !dump_opts then
            (Printf.printf "OPT: %s" s; print_newline ());
          let td = f !tdoc in
          let td = B8_typedoc.retype_tdoc td in
          if !dump_inter then begin
              Printf.printf "DUMP INTER: ["; print_newline ();
              B8_dump.dump_tdoc td;
              Printf.printf "]"; print_newline ();
            end;
          if !save_temps then begin
              let filename = Filename.temp_file s ".tmp" in
              Printf.printf "OUTPUT %s" filename; print_newline ();
              B8_emit.emit td filename;
            end;
          let new_size = B8_valid.size td in
          if !quiet then begin
              Printf.printf "OPT %s output size: %d" s new_size;
              print_newline ();
            end;
          if new_size > !tdoc_size then
            begin
              if !quiet then begin
                  Printf.printf "OPT %s output bigger document !!!" s;
                  print_newline ();
                end;              
            end
          else
          if not !valid_opts || B8_valid.validate td first_doc then
            (tdoc := td; tdoc_size := new_size)
          else
            if !quiet then begin
              Printf.printf "OPT %s output incorrect result !!!" s;
              print_newline ();
            end
        with e -> 
            if !dump_exn then
              (Printf.printf "OPT %s generate exception %s" s (
                  Printexc.to_string e); print_newline ())
) opts
*)
  else
    List.iter (fun s ->        
        if List.mem s !disabled then
          (Printf.printf "OPT: %s DISABLED" s; print_newline ())
        else
        let (f,_) = try 
            List.assoc s !B8_types.opts
          with _ ->
              Printf.printf "Unable to find OPT: %s" s;
              print_newline ();
              exit 1
        in
        if !dump_opts then
          (Printf.printf "OPT: %s" s; print_newline ());
        try 
          let td = f !tdoc in
          let td = B8_typedoc.retype_tdoc td in
          if !dump_inter then begin
              
              Printf.printf "DUMP FIRST TDOC: ["; print_newline ();
              B8_dump.dump_tdoc first_doc;
              Printf.printf "]"; print_newline ();
              Printf.printf "DUMP INTER: ["; print_newline ();
              B8_dump.dump_tdoc td;
              Printf.printf "]"; print_newline ();
            end;
          if !save_temps then begin
              let filename = Filename.temp_file s ".tmp" in
              Printf.printf "OUTPUT %s" filename; print_newline ();
              B8_emit.emit td filename;
            end;
          let new_size = B8_valid.size td in
          if !quiet then begin
              Printf.printf "OPT %s output size: %d" s new_size;
              print_newline ();
            end;
          Printf.printf "%d / %d" new_size !tdoc_size;
          print_newline ();
          if new_size > !tdoc_size then
            begin
              if !quiet then begin
                  Printf.printf "OPT %s output bigger document !!!" s;
                  print_newline ();
                end;              
            end
          else
          if not !valid_opts || B8_valid.validate td first_doc then
            (tdoc := td; tdoc_size := new_size)
          else
          if !quiet then begin
              Printf.printf "OPT %s output incorrect result !!!" s;
              print_newline ();
            end
        with e -> 
            if !dump_exn then
              (Printf.printf "OPT %s generate exception %s" s (
                  Printexc.to_string e); print_newline ())
    ) use_opts;
  let td = if not !valid_result ||
      B8_valid.validate first_doc !tdoc then
      !tdoc
    else first_doc in
  output_result td
  
